// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/18AO8/18AO8_Linux_1.x_and_2.x_DN/driver/ioctl.c $
// $Rev: 52875 $
// $Date: 2023-04-18 08:57:44 -0500 (Tue, 18 Apr 2023) $

// 18AO8: Device Driver: source file

#include "main.h"



// macros *********************************************************************

#define	AUTO_CAL_START			D28			// BCR
#define	AUTO_CAL_PASS			D29			// BCR
#define	INIT_START				D31			// BCR

#define	IRQ_AUTO_CAL_DONE		0x00000001L	// PSR



//*****************************************************************************
static void _psr_mod(dev_data_t* dev, u32 value, u32 mask)
{
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_psr_32, value & 0xFFFF, mask | 0xFFFF0000);
}



//*****************************************************************************
static int _query(dev_data_t* dev, s32* arg)
{
	switch (arg[0])
	{
		default:						arg[0]	= AO8_IOCTL_QUERY_ERROR;		break;
		case AO8_QUERY_AUTO_CAL_MS:		arg[0]	= dev->cache.auto_cal_ms;		break;
		case AO8_QUERY_CHANNEL_MASK:	arg[0]	= dev->cache.channel_mask;		break;
		case AO8_QUERY_CHANNEL_MAX:		arg[0]	= dev->cache.channels_max;		break;
		case AO8_QUERY_CHANNEL_QTY:		arg[0]	= dev->cache.channel_qty;		break;
		case AO8_QUERY_COUNT:			arg[0]	= AO8_QUERY_LAST;				break;
		case AO8_QUERY_CUSTOM_FEAT:		arg[0]	= dev->cache.custom_features;	break;
		case AO8_QUERY_DEVICE_TYPE:		arg[0]	= dev->board_type;				break;
		case AO8_QUERY_FGEN_MAX:		arg[0]	= dev->cache.fgen_max;			break;
		case AO8_QUERY_FGEN_MIN:		arg[0]	= dev->cache.fgen_min;			break;
		case AO8_QUERY_FIFO_SIZE:		arg[0]	= dev->cache.fifo_size;			break;
		case AO8_QUERY_FREF_DEFAULT:	arg[0]	= dev->cache.fref_default;		break;
		case AO8_QUERY_FSAMP_MAX:		arg[0]	= dev->cache.fsamp_max;			break;
		case AO8_QUERY_FSAMP_MIN:		arg[0]	= dev->cache.fsamp_min;			break;
		case AO8_QUERY_INIT_MS:			arg[0]	= dev->cache.initialize_ms;		break;
		case AO8_QUERY_NDIV_MAX:		arg[0]	= dev->cache.ndiv_max;			break;
		case AO8_QUERY_NDIV_MIN:		arg[0]	= dev->cache.ndiv_min;			break;
		case AO8_QUERY_SELFTEST_MS:		arg[0]	= dev->cache.selftest_ms;		break;
	}

	return(0);
}



//*****************************************************************************
int initialize_ioctl(dev_data_t* dev, void* arg)
{
	u32		bcr;
	int		i;
	long	ms	= dev->cache.initialize_ms + 5000;
	int		ret	= 0;
	int		tmp;

	// Initiate initialization.
	os_reg_mem_tx_u32(dev, dev->vaddr.gsc_bcr_32, INIT_START);

	// There is no Initialize Done interrupt so we wait manually.
	tmp	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bcr_32, INIT_START, 0);

	if (tmp)
	{
		ret	= ret ? ret : tmp;
		printf(	"%s: INITIALIZATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually check for completion.
	bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (bcr & INIT_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: INITILIZE STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Initialize the software settings.

	for (i = 0; i < DEV_IO_STREAM_QTY; i++)
	{
		if (dev->io.io_streams[i])
		{
			if (dev->io.io_streams[i]->dev_io_sw_init)
			{
				(dev->io.io_streams[i]->dev_io_sw_init)(dev, dev->io.io_streams[i]);
			}
		}
	}

	return(ret);
}



//*****************************************************************************
static int _auto_cal_start(dev_data_t* dev, void* arg)
{
	// Initiate auto-calibration.
	os_reg_mem_mx_u32(NULL, dev->vaddr.gsc_bcr_32, AUTO_CAL_START, AUTO_CAL_START);
	return(0);
}



//*****************************************************************************
static int _auto_calibrate(dev_data_t* dev, void* arg)
{
	u32			bcr;
	u32			mask;
	long		ms		= dev->cache.auto_cal_ms + 5000;
	int			ret;
	os_sem_t	sem;
	int			tmp;
	u32			value;
	gsc_wait_t	wt;

	// Safely select the Auto-Cal Done interrupt.
	mask	= IRQ_AUTO_CAL_DONE;
	value	= IRQ_AUTO_CAL_DONE;
	_psr_mod(dev, value, mask);

	// Wait for the local interrupt.
	os_sem_create(&sem);	// dummy, required for wait operations.
	memset(&wt, 0, sizeof(wt));
	wt.flags		= GSC_WAIT_FLAG_INTERNAL;
	wt.gsc			= AO8_WAIT_GSC_AUTO_CAL_DONE;
	wt.timeout_ms	= ms;
	ret				= gsc_wait_event(dev, &wt, _auto_cal_start, NULL, &sem);
	os_sem_destroy(&sem);

	if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATE IRQ TIMED OUT AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually wait for completion in case something terminates our wait early.
	tmp	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bcr_32, AUTO_CAL_START, 0);

	if (tmp)
	{
		ms	*= 2;
		ret	= ret ? ret : tmp;
		printf(	"%s: AUTOCALIBRATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (bcr & AUTO_CAL_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATION STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Final results.
	bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if ((bcr & AUTO_CAL_PASS) == 0)
	{
		ret	= ret ? ret : -EIO;
		printf(	"%s: AUTOCALIBRATION FAILED (%ld ms).\n",
				dev->model,
				ms);
	}

	return(ret);
}



//*****************************************************************************
static int _auto_cal_sts(dev_data_t* dev, s32* arg)
{
	u32	bcr;

	bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (bcr & D28)
		arg[0]	= AO8_AUTO_CAL_STS_ACTIVE;
	else if (bcr & D29)
		arg[0]	= AO8_AUTO_CAL_STS_PASS;
	else
		arg[0]	= AO8_AUTO_CAL_STS_FAIL;

	return(0);
}



//*****************************************************************************
static int _ao_config(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_AO_CONFIG_SE,
		AO8_AO_CONFIG_DIFF,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _aux_clock_mode(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_AUX_MODE_DISABLE,
		AO8_AUX_MODE_INPUT,
		AO8_AUX_MODE_OUTPUT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 1, 0);
	return(ret);
}



//*****************************************************************************
static int _aux_in_pol(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_AUX_POL_HI,
		AO8_AUX_POL_LOW,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 8, 8);
	return(ret);
}



//*****************************************************************************
static int _aux_out_pol(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_AUX_POL_HI,
		AO8_AUX_POL_LOW,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 9, 9);
	return(ret);
}



//*****************************************************************************
static int _aux_noise(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_AUX_NOISE_LOW,
		AO8_AUX_NOISE_HI,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 10, 10);
	return(ret);
}



//*****************************************************************************
static int _aux_sync_mode(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_AUX_MODE_DISABLE,
		AO8_AUX_MODE_INPUT,
		AO8_AUX_MODE_OUTPUT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 3, 2);
	return(ret);
}



//*****************************************************************************
static int _buf_clear(dev_data_t* dev, s32* arg)
{
	u32	boor;
	int	i;
	int	ret		= 0;

	// Clear the buffer.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, D11, D11);

	// Wait for the bit to clear.

	for (i = 0;; i++)
	{
		boor	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_boor_32);

		if ((boor & D11) == 0)
			break;

		if (i >= 250)
		{
			ret	= -EINVAL;
			printf(	"%s: The analog output buffer took too long to clear.\n",
					DEV_NAME);
			break;
		}
	}

	// Clear the Overflow status bit.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, 0, D15);

	// Clear the Frame Overflow status bit.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, 0, D16);

	return(ret);
}



//*****************************************************************************
static int _buf_mode(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_BUF_MODE_OPEN,
		AO8_BUF_MODE_CIRC,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 8, 8);
	return(ret);
}



//*****************************************************************************
static int _buf_over_data(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_BUF_OVER_DATA_NO,
		AO8_BUF_OVER_DATA_YES,
		-1	// terminate list
	};

	int	ret;

	if ((arg[0] == AO8_BUF_OVER_DATA_CLR) ||
		(arg[0] == AO8_BUF_OVER_DATA_CHK))
	{
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 16, 16);
	}
	else
	{
		ret	= -EINVAL;
	}

	return(ret);
}



//*****************************************************************************
static int _buf_over_frame(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_BUF_OVER_FRAME_NO,
		AO8_BUF_OVER_FRAME_YES,
		-1	// terminate list
	};

	int	ret;

	if ((arg[0] == AO8_BUF_OVER_FRAME_CLR) ||
		(arg[0] == AO8_BUF_OVER_FRAME_CHK))
	{
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 17, 17);
	}
	else
	{
		ret	= -EINVAL;
	}

	return(ret);
}



//*****************************************************************************
static int _buf_status(dev_data_t* dev, s32* arg)
{
	u32	boor;

	boor	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_boor_32);

	if (boor & D12)
		arg[0]	= AO8_BUF_STATUS_EMPTY;
	else if (boor & D15)
		arg[0]	= AO8_BUF_STATUS_FULL;
	else if (boor & D13)
		arg[0]	= AO8_BUF_STATUS_AF;
	else
		arg[0]	= AO8_BUF_STATUS_AE;

	return(0);
}



//*****************************************************************************
static int _buf_fill_level(dev_data_t* dev, s32* arg)
{
	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_range_reg(dev, arg, 0, _256K, dev->vaddr.gsc_obsr_32, 18, 0);
	return(ret);
}



//*****************************************************************************
static int _buf_thresh_level(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, _256K, dev->vaddr.gsc_obtr_32, 18, 0);
	return(ret);
}



//*****************************************************************************
static int _buf_thresh_status(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_BUF_THRESH_STATUS_CLR,
		AO8_BUF_THRESH_STATUS_SET,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_obtr_32, 20, 20);
	return(ret);
}



//*****************************************************************************
static int _burst_enable(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_BURST_ENABLE_NO,
		AO8_BURST_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 19, 19);
	return(ret);
}



//*****************************************************************************
static int _burst_ready(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_BURST_READY_NO,
		AO8_BURST_READY_YES,
		-1	// terminate list
	};

	int		ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 18, 18);
	return(ret);
}



//*****************************************************************************
static int _burst_trig_src(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_BURST_TRIG_SRC_RDG,
		AO8_BURST_TRIG_SRC_EXT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 15, 14);
	return(ret);
}



//*****************************************************************************
static int _channel_sel(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			0,
			dev->cache.channel_mask,
			dev->vaddr.gsc_ocr_32,
			7, 0);
	return(ret);
}



//*****************************************************************************
static int _clock_enable(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_CLOCK_ENABLE_NO,
		AO8_CLOCK_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 5, 5);
	return(ret);
}



//*****************************************************************************
static int _clock_ready(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_CLOCK_READY_NO,
		AO8_CLOCK_READY_YES,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 6, 6);
	return(ret);
}



//*****************************************************************************
static int _clock_src(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_CLOCK_SRC_RCG,
		AO8_CLOCK_SRC_EXT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 13, 12);
	return(ret);
}



//*****************************************************************************
static int _clock_sw(dev_data_t* dev, s32* arg)
{
	int	ret;

	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, D7, D7);
	ret	= gsc_poll_u32(dev, dev->io.tx.timeout_s * 1000, dev->vaddr.gsc_boor_32, D7, 0);
	return(ret);
}



//*****************************************************************************
static int _data_format(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_DATA_FORMAT_2S_COMP,
		AO8_DATA_FORMAT_OFF_BIN,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 25, 25);
	return(ret);
}



//*****************************************************************************
static int _dio_dir_out(dev_data_t* dev, s32* arg)
{
	u32	diopr;
	int	ret	= 0;

	switch (arg[0])
	{
		default:

			ret	= -EINVAL;
			break;

		case -1:

			diopr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_diopr_32);
			arg[0]	= diopr & (D15 | D7);
			break;

		case AO8_DIO_DIR_OUT_NONE:
		case AO8_DIO_DIR_OUT_D3_D0:
		case AO8_DIO_DIR_OUT_D7_D4:
		case AO8_DIO_DIR_OUT_D7_D0:

			os_reg_mem_mx_u32(dev, dev->vaddr.gsc_diopr_32, arg[0], D15 | D7);
			break;
	}

	return(ret);
}



//*****************************************************************************
static int _dio_read(dev_data_t* dev, s32* arg)
{
	u32	diopr;

	diopr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_diopr_32);
	arg[0]	= ((diopr & 0xF00) >> 4) | (diopr & 0xF);
	return(0);
}



//*****************************************************************************
static int _dio_write(dev_data_t* dev, s32* arg)
{
	int	ret	= 0;
	u32	v;

	if (arg[0] == -1)
	{
		ret	= _dio_read(dev, arg);
	}
	else if ((arg[0] < 0) || (arg[0] > 0xFF))
	{
		ret	= -EINVAL;
	}
	else
	{
		v	= ((arg[0] << 4) | arg[0])  & 0xF0F;
		os_reg_mem_mx_u32(dev, dev->vaddr.gsc_diopr_32, v, 0xF0F);
	}

	return(ret);
}



//*****************************************************************************
static int _gen_c_enable(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_GEN_ENABLE_NO,
		AO8_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 21, 21);
	return(ret);
}



//*****************************************************************************
static int _gen_c_ndiv(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.ndiv_min,
			dev->cache.ndiv_max,
			dev->vaddr.gsc_rcgr_32,
			23, 0);
	return(ret);
}



//*****************************************************************************
static int _gen_d_enable(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_GEN_ENABLE_NO,
		AO8_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 22, 22);
	return(ret);
}



//*****************************************************************************
static int _gen_d_ndiv(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.ndiv_min,
			dev->cache.ndiv_max,
			dev->vaddr.gsc_rdgr_32,
			23, 0);
	return(ret);
}



//*****************************************************************************
static int _irq_enable(dev_data_t* dev, s32* arg)
{
	u32	psr;
	int	ret	= 0;

	if (arg[0] == -1)
	{
		// Retrieve the current setting.
		psr		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_psr_32);
		arg[0]	= psr & AO8_IRQ_ALL;
	}
	else if ((arg[0] & AO8_IRQ_ALL) != arg[0])
	{
		ret	= -EINVAL;
	}
	else
	{
		_psr_mod(dev, arg[0], AO8_IRQ_ALL);
	}

	return(ret);
}



//*****************************************************************************
static int _load_ready(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_LOAD_READY_NO,
		AO8_LOAD_READY_YES,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 10, 10);
	return(ret);
}



//*****************************************************************************
static int _load_request(dev_data_t* dev, s32* arg)
{
	long	ms;
	int		ret;

	// Make the request.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, D9, D9);

	// Manually wait for completion.
	ms	= dev->io.tx.timeout_s * 1000;
	ret	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_boor_32, D10, D10);
	return(ret);
}



//*****************************************************************************
static int _output_mode(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_OUTPUT_MODE_SEQUENTIAL,
		AO8_OUTPUT_MODE_SIMULTANEOUS ,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 18, 18);
	return(ret);
}



//*****************************************************************************
static int _range(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_RANGE_2_5,
		AO8_RANGE_5,
		AO8_RANGE_10,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 10, 8);
	return(ret);
}



//*****************************************************************************
static int _selftest_buf_clear(dev_data_t* dev, s32* arg)
{
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, D2, D2);
	return(0);
}



//*****************************************************************************
static int _selftest_data_ready(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_SELFTEST_DATA_READY_NO,
		AO8_SELFTEST_DATA_READY_YES,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 1, 1);
	return(ret);
}



//*****************************************************************************
static int _selftest_mode(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_SELFTEST_MODE_OUT_7,
		AO8_SELFTEST_MODE_OUT_6,
		AO8_SELFTEST_MODE_OUT_5,
		AO8_SELFTEST_MODE_OUT_4,
		AO8_SELFTEST_MODE_OUT_3,
		AO8_SELFTEST_MODE_OUT_2,
		AO8_SELFTEST_MODE_OUT_1,
		AO8_SELFTEST_MODE_OUT_0,
		AO8_SELFTEST_MODE_VREF,
		AO8_SELFTEST_MODE_ZERO,
		-1	// terminate list
	};

	const s32*	list	= (dev->cache.channel_qty == 4) ? (options + 4) : options;
	int			ret;

	ret	= gsc_s32_list_reg(dev, arg, list, dev->vaddr.gsc_scr_32, 3, 0);
	return(ret);
}



//*****************************************************************************
static int _selftest_perform(dev_data_t* dev, s32* arg)
{
	long	ms;
	int		ret;

	os_time_sleep_ms(60);	// 50 ms, at least
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, D0, D0);

	// Wait for the bit to clear
	ms	= dev->cache.selftest_ms + 5000;
	ret	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bcr_32, D0, 0);
	return(ret);
}



//*****************************************************************************
static int _selftest_read(dev_data_t* dev, ao8_selftest_data_t* arg)
{
	int	i;

	for (i = 0; i < 128; i++)
		arg->data[i]	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_sdbr_32);

	return(0);
}



//*****************************************************************************
static int _trigger_sw(dev_data_t* dev, s32* arg)
{
	int	ret;

	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, D20, D20);
	ret	= gsc_poll_u32(dev, dev->io.tx.timeout_s * 1000, dev->vaddr.gsc_boor_32, D20, 0);
	return(ret);
}



//*****************************************************************************
static int _tx_io_abort(dev_data_t* dev, s32* arg)
{
	arg[0]	= gsc_write_abort_active_xfer(dev, &dev->io.tx);
	return(0);
}



//*****************************************************************************
static int _tx_io_mode(dev_data_t* dev, s32* arg)
{
	static const s32	list[]	=
	{
		GSC_IO_MODE_PIO,
		GSC_IO_MODE_BMDMA,
		-1
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, list, &dev->io.tx.io_mode);
	return(ret);
}



//*****************************************************************************
static int _tx_io_overflow_data(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_TX_IO_OVERFLOW_CHECK,
		AO8_TX_IO_OVERFLOW_IGNORE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.tx.overflow_data);
	return(ret);
}



//*****************************************************************************
static int _tx_io_overflow_frame(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_TX_IO_OVERFLOW_CHECK,
		AO8_TX_IO_OVERFLOW_IGNORE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.tx.overflow_frame);
	return(ret);
}



//*****************************************************************************
static int _tx_io_timeout(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_var(
			arg,
			AO8_IO_TIMEOUT_MIN,
			AO8_IO_TIMEOUT_INFINITE,
			&dev->io.tx.timeout_s);
	return(ret);
}



//*****************************************************************************
static int _xcvr_output(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_XCVR_OUTPUT_DISABLE,
		AO8_XCVR_OUTPUT_ENABLE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 11, 11);
	return(ret);
}



//*****************************************************************************
static int _xcvr_pol(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_XCVR_POL_DEFAULT,
		AO8_XCVR_POL_INVERT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ocr_32, 17, 17);
	return(ret);
}



//*****************************************************************************
static int _xcvr_type(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AO8_XCVR_TYPE_TTL,
		AO8_XCVR_TYPE_LVDS,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 5, 5);
	return(ret);
}



// variables ******************************************************************

const gsc_ioctl_t	dev_ioctl_list[]	=
{
	{ AO8_IOCTL_REG_READ,				(void*) gsc_reg_read_ioctl		},
	{ AO8_IOCTL_REG_WRITE,				(void*) gsc_reg_write_ioctl		},
	{ AO8_IOCTL_REG_MOD,				(void*) gsc_reg_mod_ioctl		},
	{ AO8_IOCTL_QUERY,					(void*) _query					},
	{ AO8_IOCTL_INITIALIZE,				(void*) initialize_ioctl		},
	{ AO8_IOCTL_AUTO_CALIBRATE,			(void*) _auto_calibrate			},
	{ AO8_IOCTL_AUTO_CAL_STS,			(void*) _auto_cal_sts			},
	{ AO8_IOCTL_AO_CONFIG,				(void*) _ao_config				},
	{ AO8_IOCTL_AUX_CLOCK_MODE,			(void*) _aux_clock_mode			},
	{ AO8_IOCTL_AUX_IN_POL,				(void*) _aux_in_pol				},
	{ AO8_IOCTL_AUX_OUT_POL,			(void*) _aux_out_pol			},
	{ AO8_IOCTL_AUX_NOISE,				(void*) _aux_noise				},
	{ AO8_IOCTL_AUX_SYNC_MODE,			(void*) _aux_sync_mode			},
	{ AO8_IOCTL_BUF_CLEAR,				(void*) _buf_clear				},
	{ AO8_IOCTL_BUF_MODE,				(void*) _buf_mode				},
	{ AO8_IOCTL_BUF_OVER_DATA,			(void*) _buf_over_data			},
	{ AO8_IOCTL_BUF_OVER_FRAME,			(void*) _buf_over_frame			},
	{ AO8_IOCTL_BUF_STATUS,				(void*) _buf_status				},
	{ AO8_IOCTL_BUF_FILL_LEVEL,			(void*) _buf_fill_level			},
	{ AO8_IOCTL_BUF_THRESH_LEVEL,		(void*) _buf_thresh_level		},
	{ AO8_IOCTL_BUF_THRESH_STATUS,		(void*) _buf_thresh_status		},
	{ AO8_IOCTL_BURST_ENABLE,			(void*) _burst_enable			},
	{ AO8_IOCTL_BURST_READY,			(void*) _burst_ready			},
	{ AO8_IOCTL_BURST_TRIG_SRC,			(void*) _burst_trig_src			},
	{ AO8_IOCTL_CHANNEL_SEL,			(void*) _channel_sel			},
	{ AO8_IOCTL_CLOCK_ENABLE,			(void*) _clock_enable			},
	{ AO8_IOCTL_CLOCK_READY,			(void*) _clock_ready			},
	{ AO8_IOCTL_CLOCK_SRC,				(void*) _clock_src				},
	{ AO8_IOCTL_CLOCK_SW,				(void*) _clock_sw				},
	{ AO8_IOCTL_DATA_FORMAT,			(void*) _data_format			},
	{ AO8_IOCTL_DIO_DIR_OUT,			(void*) _dio_dir_out			},
	{ AO8_IOCTL_DIO_READ,				(void*) _dio_read				},
	{ AO8_IOCTL_DIO_WRITE,				(void*) _dio_write				},
	{ AO8_IOCTL_GEN_C_ENABLE,			(void*) _gen_c_enable			},
	{ AO8_IOCTL_GEN_C_NDIV,				(void*) _gen_c_ndiv				},
	{ AO8_IOCTL_GEN_D_ENABLE,			(void*) _gen_d_enable			},
	{ AO8_IOCTL_GEN_D_NDIV,				(void*) _gen_d_ndiv				},
	{ AO8_IOCTL_IRQ_ENABLE,				(void*) _irq_enable				},
	{ AO8_IOCTL_LOAD_READY,				(void*) _load_ready				},
	{ AO8_IOCTL_LOAD_REQUEST,			(void*) _load_request			},
	{ AO8_IOCTL_OUTPUT_MODE,			(void*) _output_mode			},
	{ AO8_IOCTL_RANGE,					(void*) _range					},
	{ AO8_IOCTL_SELFTEST_BUF_CLEAR,		(void*) _selftest_buf_clear		},
	{ AO8_IOCTL_SELFTEST_DATA_READY,	(void*) _selftest_data_ready	},
	{ AO8_IOCTL_SELFTEST_MODE,			(void*) _selftest_mode			},
	{ AO8_IOCTL_SELFTEST_PERFORM,		(void*) _selftest_perform		},
	{ AO8_IOCTL_SELFTEST_READ,			(void*) _selftest_read			},
	{ AO8_IOCTL_TRIGGER_SW,				(void*) _trigger_sw				},
	{ AO8_IOCTL_TX_IO_ABORT,			(void*) _tx_io_abort			},
	{ AO8_IOCTL_TX_IO_MODE,				(void*) _tx_io_mode				},
	{ AO8_IOCTL_TX_IO_OVERFLOW_DATA,	(void*) _tx_io_overflow_data	},
	{ AO8_IOCTL_TX_IO_OVERFLOW_FRAME,	(void*) _tx_io_overflow_frame	},
	{ AO8_IOCTL_TX_IO_TIMEOUT,			(void*) _tx_io_timeout			},
	{ AO8_IOCTL_WAIT_CANCEL,			(void*) gsc_wait_cancel_ioctl	},
	{ AO8_IOCTL_WAIT_EVENT,				(void*) gsc_wait_event_ioctl	},
	{ AO8_IOCTL_WAIT_STATUS,			(void*) gsc_wait_status_ioctl	},
	{ AO8_IOCTL_XCVR_OUTPUT,			(void*) _xcvr_output			},
	{ AO8_IOCTL_XCVR_POL,				(void*) _xcvr_pol				},
	{ AO8_IOCTL_XCVR_TYPE,				(void*) _xcvr_type				},
	{ -1, NULL }
};


