// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/18AO8/18AO8_Linux_1.x_and_2.x_DN/utils/util_fsamp_ao.c $
// $Rev: 52872 $
// $Date: 2023-04-18 08:51:38 -0500 (Tue, 18 Apr 2023) $

// 18AO8: Utilities: source file

#include "main.h"



// macros *********************************************************************

#define	ABS(v)					(((v) < 0) ? (-(v)) : (v))
#define	LIMIT_MAX(v,max)		(((v) > (max)) ? (max) : (v))
#define	LIMIT_MIN(v,min)		(((v) < (min)) ? (min) : (v))
#define	LIMIT_RANGE(v,min,max)	(LIMIT_MIN((LIMIT_MAX((v),(max))),(min)))



/******************************************************************************
*
*	Function:	ao8_fsamp_ao_compute
*
*	Purpose:
*
*		Determine the best configuration for the specified output sample rate.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		fsamp	The desired sample rate.
*
*		ndiv	The Ndiv value is put here. MUST NOT BE NULL!
*
*		rate	Report the sample rate achieved here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int	ao8_fsamp_ao_compute(int fd, int index, int verbose, s32 fsamp, s32* ndiv, double* rate)
{
	double	delta;
	double	delta_best	= FLT_MAX;
	int		errs		= 0;
	s32		fref;
	double	fsamp_best	= 0;
	double	fsamp_got;
	s32		fsamp_max;	// Queried value.
	s32		fsamp_min;	// Queried value.
	double	fsamp_want;
	s32		ndiv_best	= 0;
	s32		ndiv_ll;	// lower limit
	s32		ndiv_max;	// Queried value.
	s32		ndiv_min;	// Queried value.
	s32		ndiv_ul;	// upper limit
	s32		ndiv_use;

	if (verbose)
	{
		gsc_label_index("AO Sample Rate", index);
		gsc_label_long_comma(fsamp);
		printf(" S/S");
	}

	errs	+= ao8_query(fd, -1, 0, AO8_QUERY_FREF_DEFAULT, &fref);
	errs	+= ao8_query(fd, -1, 0, AO8_QUERY_FSAMP_MAX, &fsamp_max);
	errs	+= ao8_query(fd, -1, 0, AO8_QUERY_FSAMP_MIN, &fsamp_min);
	errs	+= ao8_query(fd, -1, 0, AO8_QUERY_NDIV_MAX, &ndiv_max);
	errs	+= ao8_query(fd, -1, 0, AO8_QUERY_NDIV_MIN, &ndiv_min);

	fsamp_want	= (double) LIMIT_RANGE(fsamp, fsamp_min, fsamp_max);
	ndiv_use	= fref / fsamp_want;

	ndiv_ll		= ndiv_use - 1;
	ndiv_ul		= ndiv_use + 1;

	ndiv_ll		= (ndiv_ll < ndiv_min) ? ndiv_min : ndiv_ll;
	ndiv_ul		= (ndiv_ul > ndiv_max) ? ndiv_max : ndiv_ul;

	ndiv_ll		= (ndiv_ll > ndiv_max) ? ndiv_max : ndiv_ll;
	ndiv_ul		= (ndiv_ul < ndiv_min) ? ndiv_min : ndiv_ul;

	for (ndiv_use = ndiv_ll; ndiv_use <= ndiv_ul; ndiv_use++)
	{
		fsamp_got	= ((double) fref) / ndiv_use;
		delta		= fsamp_got - fsamp_want;

		if ((ABS(delta)) < (ABS(delta_best)))
		{
			delta_best	= delta;
			fsamp_best	= fsamp_got;
			ndiv_best	= ndiv_use;
		}
	}

	if (ndiv)
		ndiv[0]	= ndiv_best;

	if (rate)
		rate[0]	= fsamp_best;

	if (verbose)
	{
		printf("%s  (", errs ? " FAIL <---" : "");
		printf("Fref ");
		gsc_label_long_comma(fref);
		printf(", Ndiv ");
		gsc_label_long_comma(ndiv_best);
		printf(", Fsamp %.3f", fsamp_best);
		printf(")\n");
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	ao8_fsamp_get
*
*	Purpose:
*
*		Determine the output sample rate.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		verbose	Work verbosely?
*
*		get		Store the sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int	ao8_fsamp_get(int fd, int verbose, s32* get)
{
	s32		burst		= -1;
	s32		enable		= -1;
	s32		enable_c	= -1;
	int		errs		= 0;
	s32		fref;
	char	gen;
	s32		ndiv_c		= -1;
	double	rate;
	s32		set;
	s32		src			= -1;

	errs	+= ao8_query(fd, -1, 0, AO8_QUERY_FREF_DEFAULT, &fref);

	errs	+= ao8_ioctl_dsl(fd, AO8_IOCTL_CLOCK_ENABLE, &enable);
	errs	+= ao8_ioctl_dsl(fd, AO8_IOCTL_CLOCK_SRC, &src);
	errs	+= ao8_ioctl_dsl(fd, AO8_IOCTL_BURST_ENABLE, &burst);
	errs	+= ao8_ioctl_dsl(fd, AO8_IOCTL_GEN_C_ENABLE, &enable_c);
	errs	+= ao8_ioctl_dsl(fd, AO8_IOCTL_GEN_C_NDIV, &ndiv_c);

	gen		= (src == AO8_CLOCK_SRC_RCG) ? 'C' : 'D';

	if (errs)
	{
		set	= 0;

		if (verbose)
			printf("FAIL <---  (%d. service errors)\n", __LINE__);
	}
	else if (enable == AO8_CLOCK_ENABLE_NO)
	{
		set	= 0;

		if (verbose)
			printf("0  (%d. Output is disabled.)\n", __LINE__);
	}
	else if (burst == AO8_BURST_ENABLE_YES)
	{
		set	= 0;

		if (verbose)
			printf("Unknown  (%d. Bursting is enabled.)\n", __LINE__);
	}
	else if (src == AO8_CLOCK_SRC_EXT)
	{
		set	= 0;

		if (verbose)
			printf("Unknown  (%d. The clock source is external.)\n", __LINE__);
	}
	else if (enable_c == AO8_GEN_ENABLE_NO)
	{
		set	= 0;

		if (verbose)
			printf("0  (%d. Rate Generator %c is disabled.)\n", __LINE__, gen);
	}
	else
	{
		set	= fref / ndiv_c;

		if (verbose)
		{
			gsc_label_long_comma(set);
			printf(" S/S  (Ref = ");
			gsc_label_long_comma(fref);
			printf(", Source = Rate Gen %c", gen);
			printf(", Ndiv = ");
			gsc_label_long_comma(ndiv_c);
			rate	= ((double) fref) / ndiv_c;
			printf(", Fsamp = %.3f)\n", (double) rate);
		}
	}

	if (get)
		get[0]	= set;

	return(errs);
}



/******************************************************************************
*
*	Function:	ao8_fsamp_ao_report
*
*	Purpose:
*
*		Determine and report the output sample rate for the specified channel.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		chan	The index of the channel of interest.
*
*		get		Store the sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int	ao8_fsamp_ao_report(int fd, int index, int verbose, int chan, s32* get)
{
	s32		bit			= 0x1L << chan;
	char	buf[64];
	int		errs;
	s32		select		= -1;
	s32		set			= 0;

	if (verbose)
	{
		sprintf(buf, "AO Channel %d", chan);
		gsc_label_index(buf, index);
	}

	errs	= ao8_ioctl_dsl(fd, AO8_IOCTL_CHANNEL_SEL, &select);

	if (select & bit)
	{
		// The channel is enabled.
		errs	+= ao8_fsamp_get(fd, verbose, &set);
	}
	else
	{
		// The channel is disabled.
		set	= 0;

		if (verbose)
			printf("0 S/S  (Channel Disabled)\n");
	}

	if (get)
		get[0]	= set;

	return(errs);
}



/******************************************************************************
*
*	Function:	ao8_fsamp_ao_report_all
*
*	Purpose:
*
*		Determine and report the output sample rate for all channels.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		fsamp	Report the overall sample rate here.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int	ao8_fsamp_ao_report_all(int fd, int index, int verbose, s32* fsamp)
{
	s32	chans;
	int	chan;
	int	errs;
	s32	sps;
	s32	total	= 0;

	if (verbose)
		gsc_label_index("AO Sample Rates", index);

	errs	= ao8_query(fd, -1, 0, AO8_QUERY_CHANNEL_QTY, &chans);

	if (errs)
	{
		if (verbose)
			printf("FAIL <---\n");
	}
	else
	{
		if (verbose)
		{
			printf("\n");
			gsc_label_level_inc();
		}

		for (chan = 0; chan < chans; chan++)
		{
			sps		= 0;
			errs	+= ao8_fsamp_ao_report(fd, -1, verbose, chan, &sps);
			total	+= sps;
		}

		if (verbose)
		{
			gsc_label("Overall AO Rate");
			gsc_label_long_comma(total);
			printf(" S/S\n");
			gsc_label_level_dec();
		}
	}

	if (fsamp)
		fsamp[0]	= total;

	return(errs);
}


